# Copyright (c) 2011, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ACPI module."""

import bits
import _bits
import struct
from cpudetect import cpulib
from collections import namedtuple, OrderedDict
import itertools

from _bits import _acpi_get_table as get_table
from _bits import _acpi_get_table_by_index as get_table_by_index
from _bits import _acpi_install_interface as install_interface
from _bits import _acpi_objpaths as get_objpaths
from _bits import _acpi_remove_interface as remove_interface
from _bits import _acpi_terminate as terminate

def _id(v):
    return v

class TableParseException(Exception): pass

class AcpiBuffer(str):
    def __repr__(self):
        return "AcpiBuffer(" + ' '.join("{:02x}".format(ord(c)) for c in self) + ")"

    def __str__(self):
        return repr(self)

class AcpiProcessor(namedtuple('acpi_processor', ['ProcId', 'PblkAddress', 'PblkLength'])):
    __slots__ = ()

    def __repr__(self):
        return "acpi_processor(ProcId={ProcId:#x}, PblkAddress={PblkAddress:#x}, PblkLength={PblkLength})".format(**self._asdict())

ACPI_TYPE_INTEGER = 0x01
ACPI_TYPE_STRING = 0x02
ACPI_TYPE_BUFFER = 0x03
ACPI_TYPE_PACKAGE = 0x04
ACPI_TYPE_PROCESSOR = 0x0C

_acpi_object_types = {
    ACPI_TYPE_INTEGER: _id,
    ACPI_TYPE_STRING: _id,
    ACPI_TYPE_BUFFER: AcpiBuffer,
    ACPI_TYPE_PACKAGE: (lambda t: tuple(_acpi_object_to_python(v) for v in t)),
    ACPI_TYPE_PROCESSOR: (lambda args: AcpiProcessor(*args)),
}

def _acpi_object_to_python(acpi_object):
    if acpi_object is None:
        return None
    object_type, value = acpi_object
    return _acpi_object_types[object_type](value)

def _acpi_object_from_python(obj):
    if isinstance(obj, (int, long)):
        return (ACPI_TYPE_INTEGER, obj)
    # Must check AcpiBuffer before str, since AcpiBuffer derives from str
    if isinstance(obj, AcpiBuffer):
        return (ACPI_TYPE_BUFFER, obj)
    if isinstance(obj, str):
        return (ACPI_TYPE_STRING, obj)
    if isinstance(obj, AcpiProcessor):
        return (ACPI_TYPE_PROCESSOR, obj)
    # Must check tuple after any namedtuples, since namedtuples derive from tuple
    if isinstance(obj, tuple):
        return (ACPI_TYPE_PACKAGE, obj)

def evaluate(pathname, *args):
    """Evaluate an ACPI method and return the result."""
    return _acpi_object_to_python(_bits._acpi_eval(pathname, tuple(_acpi_object_from_python(arg) for arg in args)))

def scope(path):
    try:
        prefix, _ = path.rsplit('.', 1)
        return prefix
    except ValueError:
        return "/"

class DmarHardwareUnitDefinition (namedtuple('DmarHardwareUnitDefinition', ['Type', 'Length', 'Flags', 'Reserved', 'SegmentNumber', 'RegisterBaseAddress'])):
    __slots__ = ()
    def __repr__(self):
        return "DRHD( Type={Type}, Length={Length}, Flags={Flags}, Reserved={Reserved:#x}, SegmentNumber={SegmentNumber}, RegisterBaseAddress={RegisterBaseAddress:#x})".format(**self._asdict())

class DmarReservedMemoryRegionReporting (namedtuple('DmarReservedMemoryRegionReporting', ['Type', 'Length', 'Reserved', 'SegmentNumber', 'ReservedMemoryRegionBaseAddress', 'ReservedMemoryRegionLimitAddress'])):
    __slots__ = ()
    def __repr__(self):
        return "RMRR( Type={Type}, Length={Length}, Reserved={Reserved:#x}, SegmentNumber={SegmentNumber}, ReservedMemoryRegionBaseAddress={ReservedMemoryRegionBaseAddress:#x}, ReservedMemoryRegionLimitAddress={ReservedMemoryRegionLimitAddress:#x})".format(**self._asdict())

GenericReg = namedtuple('GenericRegister', ['DescriptorType', 'ResourceLength',
                        'AddressSpaceId', 'BitWidth', 'BitOffset', 'AccessSize',
                        'Address', 'EndTagByte', 'Checksum'])

FixedFuncHwReg = namedtuple('FixedFunctionalHardwareRegister', ['Type', 'VendorCode',
                            'ClassCode', 'Arg1', 'Arg0'])

class SubTable(namedtuple('SubTable', ['Type', 'Length'])):
    __slots__ = ()
    fmt_str = "<BB"
    size = struct.calcsize(fmt_str)

class MCFGResource(namedtuple('McfgResource', ['LowAddress', 'HighAddress', 'Segment', 'StartBus', 'EndBus', 'Reserved'])):
    __slots__ = ()
    fmt_str = "<LLHBBL"
    size = struct.calcsize(fmt_str)
    def __repr__(self):
        return "MCFGResource(LowAddress={LowAddress:#x}, HighAddress={HighAddress:#x}, Segment={Segment:#x}, StartBus={StartBus:#x}, EndBus={EndBus:#x}, Reserved={Reserved:#x})".format(**self._asdict())

def parse_mcfg():
    # Get a buffer containing the entire MCFG
    mcfg = get_table("MCFG")

    if mcfg is None:
        raise TableParseException('No ACPI MCFG Table found. This table is required for PCI Express.')

    # Decode the ACPI header
    size, hdr = unpack_table_header(mcfg)
    i = size # Get the size of the header we just decoded
    i += 8   # Skip the 8 byte reserved field

    if i >= len(mcfg):
        raise TableParseException('MCFG too short; reserved field truncated')

    resources_len = len(mcfg) - i
    if resources_len % MCFGResource.size:
        raise TableParseException('MCFG table size is not a multiple of the resource record size')

    # Decode each resource record
    return [MCFGResource._make(struct.unpack_from(MCFGResource.fmt_str, mcfg, offset))
            for offset in range(i, len(mcfg), MCFGResource.size)]

def parse_dmar(printflag=False):
    """Parse and optionally print an DMAR."""

    ACPI_DMAR_TYPE_DRHD = 0
    ACPI_DMAR_TYPE_RMRR = 1
    ACPI_DMAR_TYPE_ATSR = 2
    ACPI_DMAR_TYPE_RHSA = 3

    # Get a buffer containing the entire DMAR
    dmar = get_table("DMAR")

    if dmar == None:
        if printflag:
            print 'No ACPI DMAR Table found!'
        return

    # Decode the ACPI header
    size, hdr = unpack_table_header(dmar, 0, printflag)
    i = size # Get the size of the header we just decoded

    # Decode the host address width field
    fmt_str = "<BB10s"
    HostAddress = namedtuple('HostAddress', ['width', 'Flags', 'Reserved', ])
    ha = HostAddress._make( struct.unpack_from(fmt_str, dmar, i) )
    if printflag:
        print 'Host Address Width = {0} ({1} address bits)'.format(ha.width, ha.width + 1)
    i += struct.calcsize(fmt_str)

    # DmarHardwareUnitDefinition = namedtuple('DmarHardwareUnitDefinition', ['Type', 'Length', 'Flags', 'Reserved', 'SegmentNumber', 'RegisterBaseAddress'])
    #DmarReservedMemoryRegionReporting = namedtuple('DmarReservedMemoryRegionReporting', ['Type', 'Length', 'Reserved', 'SegmentNumber', 'ReservedMemoryRegionBaseAddress', 'ReservedMemoryRegionLimitAddress'])
    DmarRootPortATSCapabilityReporting = namedtuple('DmarRootPortATSCapabilityReporting', ['Type', 'Length', 'Flags', 'Reserved', 'SegmentNumber'])
    DmarRemappingHardwareStatusAffinity = namedtuple('DmarRemappingHardwareStatusAffinity', ['Type', 'Length', 'Reserved', 'RegisterBaseaAddress', 'ProximityDomain'])
    DmarDeviceScope = namedtuple('DeviceScope', ['Type', 'Length', 'Reserved', 'EnumerationID', 'StartBusNumber'])
    DmarDeviceScopePath = namedtuple('DeviceScopePath', ['PciDevice', 'PciFunction'])

    def decodeDeviceScope(start, current, end):
        while current < end:
            fmt_str = "<BBHBB"
            dscope = DmarDeviceScope._make( struct.unpack_from(fmt_str, dmar, start + current) )
            if printflag: print dscope
            k = struct.calcsize(fmt_str)
            while k < dscope.Length:
                fmt_str = "<BB"
                path = DmarDeviceScopePath._make( struct.unpack_from(fmt_str, dmar, start + current + k) )
                if printflag: print path
                k += struct.calcsize(fmt_str)
            current += dscope.Length

    # Decode each subtable
    while i != len(dmar):
        subtable = SubTable._make(struct.unpack_from(SubTable.fmt_str, dmar, i))
        # if printflag: print subtable
        if subtable.Type == ACPI_DMAR_TYPE_DRHD:
            fmt_str = "<HHBBHQ"
            drhd = DmarHardwareUnitDefinition._make( struct.unpack_from(fmt_str, dmar, i) )
            if printflag: print drhd
            decodeDeviceScope( i, struct.calcsize(fmt_str), subtable.Length)
        if subtable.Type == ACPI_DMAR_TYPE_RMRR:
            fmt_str = "<HHHHQQ"
            rmrr = DmarReservedMemoryRegionReporting._make( struct.unpack_from(fmt_str, dmar, i) )
            if printflag: print rmrr
            decodeDeviceScope( i, struct.calcsize(fmt_str), subtable.Length)
        if subtable.Type == ACPI_DMAR_TYPE_ATSR:
            fmt_str = "<HHBBH"
            atsr = DmarRootPortATSCapabilityReporting._make( struct.unpack_from(fmt_str, dmar, i) )
            if printflag: print atsr
            decodeDeviceScope( i, struct.calcsize(fmt_str), subtable.Length)
        if subtable.Type == ACPI_DMAR_TYPE_RHSA:
            fmt_str = "<HHIQI"
            rhsa = DmarRemappingHardwareStatusAffinity._make( struct.unpack_from(fmt_str, dmar, i) )
            if printflag: print rhsa
            decodeDeviceScope( i, struct.calcsize(fmt_str), subtable.Length)
        i += subtable.Length

class ApicInfo (namedtuple('ApicInfo', ['LocalApicAddress', 'Flags'])):
    __slots__ = ()
    def __repr__(self):
        return "ApicInfo( LocalApicAddress={LocalApicAddress:#x}, Flags={Flags:#x})".format(**self._asdict())

class MadtLocalApic(namedtuple('MadtLocalApic', ['Type', 'Length', 'ProcId', 'ApicId', 'Flags'])):
    __slots__ = ()
    fmt_str = "<BBBBI"
    size = struct.calcsize(fmt_str)
    def __repr__(self):
        return "MadtLocalApic(Type={Type}, Length={Length}, ProcId={ProcId:#x}, ApicId={ApicId:#x}, Flags={Flags:#x})".format(**self._asdict())

class MadtLocalx2Apic(namedtuple('MadtLocalx2Apic', ['Type', 'Length', 'Reserved', 'X2ApicId', 'Flags', 'UID'])):
    __slots__ = ()
    fmt_str = "<BBHIII"
    size = struct.calcsize(fmt_str)
    def __repr__(self):
        return "MadtLocalApic(Type={Type}, Length={Length}, Reserved={Reserved:#x}, X2ApicId={X2ApicId:#x}, Flags={Flags:#x}, UID={UID})".format(**self._asdict())

MADT_ENABLED = 1
MADT_TYPE_LOCAL_APIC = 0
MADT_TYPE_LOCAL_X2APIC = 9

def parse_madt(printflag=False):
    """Parse and optionally print an MADT."""
    procid_apicid = {}
    uid_x2apicid = {}

    # Get a buffer containing the entire MADT
    madt = get_table("APIC")

    if madt is None:
        return None, None

    # Decode the ACPI header
    size, hdr = unpack_table_header(madt, 0, printflag)
    i = size # Get the size of the header we just decoded

    # Decode the local APIC address and APIC flags fields
    # ApicInfo = namedtuple('ApicInfo', ['LocalApicAddress', 'Flags'])
    fmt_str = "<II"
    apic = ApicInfo._make( struct.unpack_from(fmt_str, madt, i) )
    if printflag:
        print apic
    i += struct.calcsize(fmt_str)

    # Decode each subtable
    while i != len(madt):
        subtable = SubTable._make(struct.unpack_from(SubTable.fmt_str, madt, i))
        if subtable.Type == MADT_TYPE_LOCAL_APIC:
            localApic = MadtLocalApic._make(struct.unpack_from(MadtLocalApic.fmt_str, madt, i))
            if printflag:
                print localApic
            if localApic.Flags & MADT_ENABLED:
                procid_apicid[localApic.ProcId] = localApic.ApicId
        if subtable.Type == MADT_TYPE_LOCAL_X2APIC:
            x2Apic = MadtLocalx2Apic._make(struct.unpack_from(MadtLocalx2Apic.fmt_str, madt, i))
            if printflag:
                print x2Apic
            if x2Apic.Flags & MADT_ENABLED:
                uid_x2apicid[x2Apic.UID] = x2Apic.X2ApicId
        i += subtable.Length
    if printflag:
        print 'ProcID to ApicID = {0}'.format(procid_apicid)
        print 'UID to X2ApicID = {0}'.format(uid_x2apicid)
    return (procid_apicid, uid_x2apicid)

class TableHeader (namedtuple('TableHeader', ['signature', 'length', 'revision',
                         'checksum', 'oemid', 'oemtableid', 'oemrevision',
                         'creatorid', 'creatorrevision'])):
    __slots__ = ()
    fmt_str = "<4sIBB6s8sI4sI"
    size = struct.calcsize(fmt_str)

    def __repr__(self):
        return "TableHeader( signature={signature!r}, length={length}, revision={revision}, checksum={checksum:#x}, oemid={oemid!r}, oemtableid={oemtableid!r}, oemrevision={oemrevision:#x}, creatorid={creatorid!r}, creatorrevision={creatorrevision:#x})".format(**self._asdict())

    def repack(self):
        return struct.pack(self.fmt_str, *self)

def unpack_table_header(tbl, offset=0, printflag=False):
    """Parse, optionally print and return an unpacked ACPI header"""
    hdrsize = struct.calcsize(TableHeader.fmt_str)
    if offset + hdrsize > len(tbl):
        raise TableParseException('ACPI table header truncated; has {} bytes but expected {} bytes'.format(len(tbl) - offset, hdrsize))
    hdr = TableHeader._make( struct.unpack_from(TableHeader.fmt_str, tbl, offset) )
    if printflag:
        print hdr
    return (hdrsize, hdr)

def get_cpupaths(*args):
    cpupaths = _bits._acpi_cpupaths(*args)
    procid_apicid, uid_x2apicid = parse_madt()
    if procid_apicid is None or uid_x2apicid is None:
        return cpupaths
    enabled_cpupaths = []
    for cpupath in cpupaths:
        procdef = evaluate(cpupath)
        uid = evaluate(cpupath + "._UID")
        if (procdef is not None and procdef.ProcId in procid_apicid) or (uid is not None and uid in uid_x2apicid):
           enabled_cpupaths.append(cpupath)
    return enabled_cpupaths

def find_procid():
    cpupaths = get_cpupaths()
    cpupath_procid = {}
    for cpupath in cpupaths:
        cpupath_procid[cpupath] = evaluate(cpupath).ProcId
    return OrderedDict(sorted(cpupath_procid.items()))

def find_uid():
    cpupaths = get_cpupaths()
    cpupath_uid = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "._UID")
        cpupath_uid[cpupath] = value
    return OrderedDict(sorted(cpupath_uid.items()))

def commonprefix(l):
    """Return the common prefix of a list of strings."""
    if not l:
        return ''
    prefix = l[0]
    for s in l[1:]:
        for i, c in enumerate(prefix):
            if c != s[i]:
                prefix = s[:i]
                break
    return prefix

def factor_commonprefix(l):
    if not l:
        return ''
    if len(l) == 1:
        return l[0]
    prefix = commonprefix(l)
    prefixlen = len(prefix)
    return prefix + "{" + ", ".join([s[prefixlen:] for s in l]) + "}"

def display_cpupaths():
    """Find and display ACPI processor name space paths"""
    cpupaths = get_cpupaths()
    print factor_commonprefix(cpupaths)

def display_cpu_info():
    cpupaths = get_cpupaths()
    cpupath_procid = find_procid()
    cpupath_uid = find_uid()
    procid_apicid, uid_x2apicid = parse_madt()
    socketindex_cpuscope = {}
    for cpupath in cpupaths:
        print cpupath,
        def socket_str(apicid):
            socket_index = bits.socket_index(apicid)
            if socket_index is None:
                return ''
            return ', socket index = 0x%02x' % socket_index
        def apicid_str(apicid):
            if apicid is None:
                return 'no APIC ID'
            return 'APIC ID 0x%02x%s' % (apicid, socket_str(apicid))
        procid = cpupath_procid.get(cpupath, None)
        if procid is not None:
            print 'ProcID = %-2u (%s)' % (procid, apicid_str(procid_apicid.get(procid, None))),
        uid = cpupath_uid.get(cpupath, None)
        if uid is not None:
            print '_UID = %s (%s)' % (uid, apicid_str(uid_x2apicid.get(uid, None))),
        print
        socketindex_cpuscope.setdefault(bits.socket_index(procid_apicid.get(procid, None)), []).append(scope(cpupath))
    for value, scopes in socketindex_cpuscope.iteritems():
        unique_scopes = set(scopes)
        print 'socket {0} contains {1} processors and {2} ACPI scope: {3}'.format(value, len(scopes), len(unique_scopes), ','.join(sorted(unique_scopes)))

def display_acpi_method(method, print_one):
    """Helper function that performs all basic processing for evaluating an ACPI method"""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "." + method)
        uniques.setdefault(value, []).append(cpupath)

    print "%u unique %s values" % (len(uniques), method)
    for value, cpupaths in sorted(uniques.iteritems(), key=(lambda (k,v): v)):
        print
        print factor_commonprefix(cpupaths)
        if value is None:
            print "No %s found for these CPUs" % method
        else:
            print_one(value)

def unpack_register(reg):
    """Unpack a register string to (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)"""
    register = GenericReg._make( struct.unpack("<BHBBBBQBB", reg) )
    return register

def display_csd():
    """Find and display _CSD (C-State Dependency) methods"""
    def print_csd(csd):
        if not isinstance(csd, tuple):
            print "No _CSD found for these CPUs"
            return
        fmt_str = "              %-10s %-10s %-8s %-10s %-9s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcs", "Index")
        for csd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors, Index] in enumerate(csd):
            fmt_str = "Dependency[%u] 0x%-8x 0x%-8x 0x%-6x 0x%-8x 0x%-7x %-8x"
            print fmt_str % (csd_num, NumEntries, Revision, Domain, CoordType, NumProcessors, Index)
    display_acpi_method("_CSD", print_csd)

def makeFFH(r):
    return FixedFuncHwReg(Type=r.AddressSpaceId, VendorCode=r.BitWidth,
                          ClassCode=r.BitOffset, Arg1=r.AccessSize,
                          Arg0=r.Address)

class CState(namedtuple('CState', ['Register', 'Type', 'Latency', 'Power'])):
    __slots__ = ()
    def __repr__(self):
        r = self.Register
        if r.AddressSpaceId == 0x7f:
            r = makeFFH(self.Register)
        return "Cstate(Register={reg}, Type={Type}, Latency={Latency}, Power={Power})".format(reg = r, **self._asdict())

def parse_cst():
    """Parse the _CST for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        pkg = evaluate(cpupath + "._CST")
        if pkg is not None:
            cstate_count = pkg[0]
            cstates = tuple( CState(unpack_register(subpkg[0]), subpkg[1], subpkg[2], subpkg[3]) for subpkg in pkg[1:] )
            if cstate_count != len(cstates):
                raise TableParseException('_CST parse error: State count = {} but table had {} entries'.format(cstate_count, len(cstates)))
        else:
            cstates = None
        uniques.setdefault(cstates, []).append(cpupath)
    return uniques

def print_cst(cstates):
    if not cstates:
        print "Empty _CST was returned."
        return
    print "    {:40s}  {:7s}  {:12s}  {:10s}".format("Description", "OS Type", "Latency (us)", "Power (mW)")
    for cstate_num, cstate_data in enumerate(cstates):
        desc = ''
        if cstate_data.Register.AddressSpaceId == 0x7f:
            ffh = makeFFH(cstate_data.Register)
            if ffh.VendorCode == 0:
                desc += "C1 Halt"
            elif (ffh.VendorCode == 1) and (ffh.ClassCode == 1):
                desc += "C1 I/O then Halt I/O port address = {#x}".format(Register.Arg0)
            elif (ffh.VendorCode == 1) and (ffh.ClassCode == 2):
                desc += "MWAIT {:#02x} ({})".format(ffh.Arg0, cpulib.mwait_hint_to_cstate(ffh.Arg0))
                desc += " {} {}BMAvoid".format(("SWCoord", "HWCoord")[bool(ffh.Arg1 & 1)], ("!", "")[bool(ffh.Arg1 & (1 << 1))])
        print "C{cstate_num:<d}  {desc:40.40s}  ACPI C{Type:<1d}  {Latency:<12d}  {Power:<10d}".format(cstate_num=cstate_num, desc=desc, **CState(*cstate_data)._asdict())

def display_cst():
    """Find and display _CST (C-States) methods"""
    def print_cst(cst):
        for (reg, cst_type, latency, power) in cst[1:]:
            r = unpack_register(reg)
            print
            if r.AddressSpaceId == 0x7f:
                # Decode register as FFH
                print "Fixed Functional Hardware (FFH) detected, using Intel-specific decode"
                ffh = FixedFuncHwReg(Type=r.AddressSpaceId, VendorCode=r.BitWidth,
                                     ClassCode=r.BitOffset, Arg1=r.AccessSize,
                                     Arg0=r.Address)
                fmt_str = "%-5s %-10s %-9s %-10s %-8s"
                print fmt_str % ("Type", "VendorCode", "ClassCode", "Arg0", "Arg1")
                fmt_str = "0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (ffh.Type, ffh.VendorCode, ffh.ClassCode, ffh.Arg0, ffh.Arg1)
                if ffh.VendorCode == 0:
                    print "  C1 Halt"
                elif (ffh.VendorCode == 1) and (ffh.ClassCode == 1):
                    print "  C1 I/O then Halt I/O port address = 0x%x" % Arg0
                elif (ffh.VendorCode == 1) and (ffh.ClassCode == 2):
                    fmt_str = "  Native C-state instruction, Hint: 0x%x (%s)"
                    print fmt_str % (ffh.Arg0, cpulib.mwait_hint_to_cstate(ffh.Arg0))
                    fmt_str = "  HW coordination=%u, Bus Master avoidance=%u"
                    print fmt_str % (bool(ffh.Arg1 & 1), bool((ffh.Arg1 & (1 << 1))))
            else:
                # Decode register as actual hardware resource
                fmt_str = "%-11s %-10s %-9s %-10s %-8s"
                print fmt_str % ("AddrSpaceId", "BitWidth", "BitOffset", "AccessSize", "Address")
                fmt_str = "0x%-9x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
            # Decode and print ACPI c-state, latency, & power
            fmt_str = "%-7s %-7s %-7s"
            print fmt_str % ("C-state", "Latency", "Power")
            cst_type_str = "ACPI C%u" % cst_type
            latency_str = "%uus" % latency
            power_str = "%umW" % power
            print fmt_str % (cst_type_str, latency_str, power_str)
    display_acpi_method("_CST", print_cst)

def display_pct():
    """Find and display _PCT (Performance Control) methods"""
    def print_pct(pct):
        fmt_str = "%-12s %-5s %-10s %-9s %-10s %-8s"
        print fmt_str % ("", "Type",        "VendorCode","ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(pct):
            r = unpack_register(reg)
            fmt_str = "%-12s 0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
            print fmt_str % (reg_type[reg_num], r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
    display_acpi_method("_PCT", print_pct)

def display_pdl():
    """Find and display _PDL (P-State Depth Limit) methods"""
    def print_pdl(pdl):
        print "Lowest performance state that OSPM can use = %u" % pdl
    display_acpi_method("_PDL", print_pdl)

def display_ppc():
    """Find and display _PPC (Performance Present Capabilities) methods"""
    def print_ppc(ppc):
        print "Highest performance state that OSPM can use = %u" % ppc
    display_acpi_method("_PPC", print_ppc)

def display_psd():
    """Find and display _PSD (P-State Dependency) methods"""
    def print_psd(psd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(psd, tuple):
            print "No _PSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcessors")
        for psd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors] in enumerate(psd):
            fmt_str = "Dependency[%u]: 0x%-8x 0x%-8x 0x%-8x 0x%-8x 0x%-8x"
            print fmt_str % (psd_num, NumEntries, Revision, Domain, CoordType, NumProcessors)
    display_acpi_method("_PSD", print_psd)

class PSD(namedtuple('PSD', ['NumEntries', 'Revision', 'Domain', 'CoordType', 'NumProcessors'])):
    __slots__ = ()
    def __repr__(self):
        return "_PSD(NumEntries={NumEntries}, Revision={Revision}, Domain={Domain:#x}, CoordType={CoordType:#x}, NumProcessors={NumProcessors})".format(**self._asdict())

def parse_psd():
    """Parse the _PSD for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "._PSD")
        if value is not None:
            value = PSD(*(value)[0])
        uniques.setdefault(value, []).append(cpupath)
    return uniques

class PState(namedtuple('PState', ['CoreFrequency', 'Power', 'Latency', 'BusMasterLatency', 'Control', 'Status'])):
    __slots__ = ()
    def __repr__(self):
        return "Pstate(CoreFrequency={CoreFrequency} MHz, Power={Power} mW, Latency={Latency} us, BusMasterLatency={BusMasterLatency} us, Control={Control:#x}, Status={Status:#x})".format(**self._asdict())

def parse_pss():
    """Parse the _PSS for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        pkg = evaluate(cpupath + "._PSS")
        if pkg is not None:
            pstates = tuple(PState(*subpkg) for subpkg in pkg)
        else:
            pstates = None
        uniques.setdefault(pstates, []).append(cpupath)
    return uniques

def show_pss(pstates):
    if not pstates:
        return "Empty _PSS"
    lines = ["    {:10s} {:10s} {:12s} {:14s} {:10s} {:10s}".format("Freq (MHz)", "Power (mW)", "Latency (us)", "BMLatency (us)", "Control", "Status")]
    for pstate_num, pstate_data in enumerate(pstates):
        lines.append("P{pstate_num:<2d} {CoreFrequency:<10d} {Power:<10d} {Latency:<12d} {BusMasterLatency:<14d} {Control:#010x} {Status:#010x}".format(pstate_num=pstate_num, **PState(*pstate_data)._asdict()))
    return lines

def print_pss(pstates):
    print "\n".join(show_pss(pstates))

def display_pss():
    """Find and display _PSS (Performance Supported States) methods"""
    display_acpi_method("_PSS", print_pss)

def display_ptc():
    """Find and display _PTC (Processor Throttling Control) methods"""
    def print_ptc(ptc):
        fmt_str = "%-12s %-13s %-12s %-11s %-12s %s"
        print fmt_str % ("Processor",  "AddrSpaceId", "BitWidth",   "BitOffset", "AccessSize", "Address")
        print fmt_str % ("Throttling", "Type",        "VendorCode", "ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(ptc):
            r = unpack_register(reg)
            fmt_str = "%-12s 0x%-11x 0x%-10x 0x%-9x 0x%-10x 0x%x"
            print fmt_str % (reg_type[reg_num], r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
    display_acpi_method("_PTC", print_ptc)

def display_tdl():
    """Find and display _TDL (T-State Depth Limit) methods"""
    def print_tdl(tdl):
        print "Lowest power throttling state that OSPM may use = %u" % tdl
    display_acpi_method("_TDL", print_tdl)

def display_tpc():
    """Find and display _TPC (Throttling Present Capabilities) methods"""
    def print_tpc(tpc):
        print "Highest throttling state that OSPM can use = %u" % tpc
    display_acpi_method("_TPC", print_tpc)

def display_tsd():
    """Find and display _TSD (T-State Dependency) methods"""
    def print_tsd(tsd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tsd, tuple):
            print "No _TSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s    %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcessors")
        for tsd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors] in enumerate(tsd):
            fmt_str = "Dependency[%u]: 0x%-8u 0x%-8u 0x%-8x 0x%-8x    %-8u"
            print fmt_str % (tsd_num, NumEntries, Revision, Domain, CoordType, NumProcessors)
    display_acpi_method("_TSD", print_tsd)

def display_tss():
    """Find and display _TSS (Throttling Supported States) methods"""
    def print_tss(tss):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tss, tuple):
            print "No _TSS found for these CPUs"
            return
        fmt_str = "    %-9s %-10s %-10s %-10s %-10s"
        print fmt_str % ("Percent", "Power", "Latency", "Control", "Status")
        for tstate_num, [percent, power, latency, control, status] in enumerate(tss):
            fmt_str = "T%-2u %6u%%   0x%08x 0x%08x 0x%08x 0x%08x"
            print fmt_str % (tstate_num, percent, power, latency, control, status)
    display_acpi_method("_TSS", print_tss)

def display_uid():
    """Find and display _UID"""
    def print_uid(uid):
        print "_UID = %s" % uid
    display_acpi_method("_UID", print_uid)

def display_mcfg():
    """Find and display MCFG (PCIE memory-mapped configuration space base address) table"""
    try:
        resources = parse_mcfg()
    except TableParseException as e:
        print "No valid MCFG description table exists:"
        print e
        return
    for i, resource in enumerate(resources):
        print 'Resource Descriptor {}: {}'.format(i, resource)

def get_table_list(printflag=False):
    """Get the list of ACPI table signatures"""
    tables = itertools.takewhile(lambda table: table is not None,
                                 (get_table_by_index(index) for index in itertools.count()))
    signatures = [table[:4] for table in tables]
    if printflag:
        print signatures
    return signatures

def dump(name=""):
    for path in get_objpaths(name):
        print '{} : {!r}'.format(path, evaluate(path))
